# AutoGive (with Unit/Resource Donation)
# by falcontx
#
# Inspired by:
#  - One Key Unit Donation by jiver
#  - LazyShare by SvartfaR
#
# Compatibility:
#  - One Key Unit Donation: incompatible
#      The unit and resource donation functionality provided by this mod is very
#      similar to what is provided by One Key Unit Donation. Running both would
#      be redundant, so One Key Unit Donation is considered incompatible.
#  - LazyShare: compatible
#      The way that allies are identified by LazyShare is unique in that players are
#      assigned an identifier that is indicated on the scoreboard. Since some
#      users may prefer this over the static user assignments that this script
#      uses, this mod has been made compatible with LazyShare.


# Normally, you should not give to an AI player, as they will not know how to
# properly handle the additional units. However, the ability to do this is
# useful for testing.
local allowGiveToAI = false

function UpdateScoreData(newData) 
  scoreData = table.deepcopy(newData) 
end 

local function GetScoreData() 
  return scoreData 
end

function giveSelectedUnits(ally, giveAll)
  local scoreData = GetScoreData()
  local allies = {}
  local alliesexist = false
  for index, playerInfo in GetArmiesTable().armiesTable do
    if (not playerInfo.civilian and index != GetFocusArmy() and IsAlly(GetFocusArmy(), index) and (playerInfo.human or allowGiveToAI)) then
      playerInfo.index = index
      playerInfo.score = scoreData[index].general.score - scoreData[index].units.cdr.kills * 134000
      table.insert(allies, playerInfo)
      if not playerInfo.outOfGame then
        alliesexist = true
      end
    end
  end
  if alliesexist then
    if giveAll then
      UISelectionByCategory("ALLUNITS", false, false, false, false)
      if allies[ally] and not allies[ally].outOfGame then
        SimCallback({Func="GiveUnitsToPlayer", Args={ From=GetFocusArmy(), To=allies[ally].index},} , true)
        SimCallback({Func="GiveResourcesToPlayer", Args={ From=GetFocusArmy(), To=allies[ally].index, Mass=1.0, Energy=1.0},} , true)
      end
      table.sort(allies, function(a,b) return a.score > b.score end)
      for index, playerInfo in allies do
        if not playerInfo.outOfGame then
          SimCallback({Func="GiveUnitsToPlayer", Args={ From=GetFocusArmy(), To=playerInfo.index},} , true)
          SimCallback({Func="GiveResourcesToPlayer", Args={ From=GetFocusArmy(), To=playerInfo.index, Mass=1.0, Energy=1.0},} , true)
        end
      end
      return true
    elseif allies[ally] and not allies[ally].outOfGame then
      SimCallback({Func="GiveUnitsToPlayer", Args={ From=GetFocusArmy(), To=allies[ally].index},} , true)
    end
  else
    return false
  end
end

function giveResources(ally, type)
  local amount = 50.0
  local allies = {}
  for index, playerInfo in GetArmiesTable().armiesTable do
    if (not playerInfo.civilian and index ~= GetFocusArmy() and IsAlly(GetFocusArmy(), index) and (playerInfo.human or allowGiveToAI)) then
      playerInfo.index = index
      table.insert(allies, playerInfo)
    end
  end
  if allies[ally] and not allies[ally].outOfGame then
    if type == "energy" then
      SimCallback({Func="GiveResourcesToPlayer", Args={ From=GetFocusArmy(), To=allies[ally].index, Mass=0.0, Energy=amount/100.0},} , true)
    else
      SimCallback({Func="GiveResourcesToPlayer", Args={ From=GetFocusArmy(), To=allies[ally].index, Mass=amount/100.0, Energy=0.0},} , true)
    end
  end
end

local lazyshare_version = 0 -- default is no lazyshare; detection script updates this

function commanderWatch(commander, ally, thread)
  local prevProgress = 0
  local cHealth = commander:GetHealth()
  while not commander:IsDead() and cHealth > 0 do
    WaitSeconds(0.1)
    if commander:GetHealth() < cHealth - 500 then
      cHealth = math.floor(commander:GetHealth())
      print("Commander Health: " .. cHealth .. " / " .. math.floor(commander:GetMaxHealth()))
    elseif commander:GetHealth() > cHealth + 50 then
      cHealth = commander:GetHealth()
    end
  end
  if giveSelectedUnits(ally, true) then
    print("Commander died; gave units to ally.")
  end
  KillThread(thread)
  thread = nil
end

function assassinationAutoGive(ally)
  if SessionGetScenarioInfo().Options.Victory == "demoralization" then
    local alliesexist = false
    for i, aI in GetArmiesTable().armiesTable do
      if (allowGiveToAI or aI.human) and GetFocusArmy() != i and IsAlly(GetFocusArmy(), i) then
        alliesexist = true
      end
    end
    local currentSelection = GetSelectedUnits() or {}
    ConExecute('UI_SelectByCategory +nearest COMMAND')
    local selection = GetSelectedUnits()
    SelectUnits(currentSelection)
    commander = selection[1]
    if commander.AutoWatchThread then
      KillThread(commander.AutoWatchThread)
      commander.AutoWatchThread = nil
    end
    if commander.WatchThread then
      print("AutoGive DISABLED")
      KillThread(commander.WatchThread)
      commander.WatchThread = nil
    elseif alliesexist then
      if lazyshare_version > 1 then
        local lazyshareIDs = import('/mods/lazyshare/mimc_diplomacy.lua').getLazyShareIDs()
        if lazyshareIDs[ally] != nil then
          local armiesTable = GetArmiesTable().armiesTable
          message = "Will give units/resources to " .. armiesTable[lazyshareIDs[ally]].nickname .. " if they are still in the game."
          local acount = 1
          for index, playerInfo in GetArmiesTable().armiesTable do
            if acount > 0 and (not playerInfo.civilian and index != GetFocusArmy() and IsAlly(GetFocusArmy(), index) and (playerInfo.human or allowGiveToAI)) then
              if lazyshareIDs[ally] == index then
                ally = acount
                acount = -1
              end
              acount = acount + 1
            end
          end
        else
          return false
        end
      elseif lazyshare_version == 1 then
        message = "Will give units/resources to the ally with the highest score."
        ally = -1
      else
        local allies = {}
        for index, playerInfo in GetArmiesTable().armiesTable do
          if (not playerInfo.civilian and index != GetFocusArmy() and IsAlly(GetFocusArmy(), index) and (playerInfo.human or allowGiveToAI)) then
            playerInfo.index = index
            table.insert(allies, playerInfo)
          end
        end

        if allies[ally] and not allies[ally].outOfGame then
          message = "Will give units/resources to " .. allies[ally].nickname .. ", if they are still in the game."
        else
          print("AutoGive can not be enabled for this ally.")
          return false
        end
      end
    
      print("AutoGive ENABLED; " .. message)
      commander.WatchThread = ForkThread(commanderWatch, commander, ally, commander.WatchThread)
    else
      print("AutoGive can not be enabled; No allies exist.")
    end
  else
    print("AutoGive can not be enabled; Victory condition is not Assassination.")
  end
end

function Init()
  local alliesexist = false
  local disabledefault = false
  LOG("AutoGive: Init()")

  for i, mod in __active_mods do
    if mod["uid"] == "c31251c0-8199-11dd-ad8b-0800200c9a66" then
      lazyshare_version = mod["version"]
      LOG("AutoGive: detected lazyshare v" .. lazyshare_version)
    end
    if mod["uid"] == "7c892c84-1ffc-402d-b328-6579bea5f1f8" or mod["uid"] == "516bdf61-c79a-4625-a758-60d1ba58a497" or mod["uid"] == "926bdf61-c79a-4625-a758-60d1b251a417" then
      disabledefault = true
      LOG("AutoGive: detected Phantom mod; AutoGive will not be enabled by default")
    end
  end

	import('/mods/AutoGive/keymapping.lua').initUserKeyActions()

  if not disabledefault and not SessionIsReplay() and GetFocusArmy() != -1 and SessionGetScenarioInfo().Options.Victory == "demoralization" then
    for i, aI in GetArmiesTable().armiesTable do
      if (allowGiveToAI or aI.human) and GetFocusArmy() != i and IsAlly(GetFocusArmy(), i) then
        alliesexist = true
      end
    end
    if alliesexist then
      ConExecute('UI_SelectByCategory +nearest COMMAND')
      local selection = GetSelectedUnits()
      SelectUnits(nil) 
      local commander = selection[1]
      print("Assassination victory condition detected; AutoGive ENABLED; Will give units/resources to the ally with highest score.")
      commander.AutoWatchThread = ForkThread(commanderWatch, commander, -1, commander.AutoWatchThread)
    end
  end
end
